-module(server).
-export([main_server/0,main_genserver/0]).

% If we don't want to keep track of all PIDs, we can bind a PID to an atom like
% this. We can then use it when calling the process.
start_server() ->
    Pid = spawn(fun server/0),
    catch unregister(server),
    register(server, Pid).


% Initial math server, can multiply, add and subtract, square
server() ->
    receive
       { mul, X, Y, From, Ref } ->
            spawn(fun() -> Val = X*Y, From ! { answer, X, Y, mul, Val, Ref } end );
       { add, X, Y, From, Ref } ->
            spawn(fun() -> Val = X+Y, From ! { answer, X,Y,add,Val, Ref } end );
       { sub, X, Y, From, Ref } ->
            spawn(fun() -> Val = X-Y, From ! { answer, X,Y, sub, Val, Ref } end );
       { square, X, From, Ref } ->
            spawn(fun() -> Val = X*X, From ! { answer, Val, Ref } end );
       _ -> ok
     end,
     server().

start_genserver() ->
    genserver:start(server,[], fun loop_function/2).
    

loop_function(Requests, { Op, V1, V2 }) ->
    case search({Op,V1,V2}, Requests) of
       false ->
          case Op of 
            mul -> { reply, V1*V2, [{Op, V1, V2, V1*V2} | Requests] };
            sub -> { reply, V1-V2, [{Op, V1, V2, V1-V2} | Requests] };
            add -> { reply, V1+V2, [{Op, V1, V2, V1+V2} | Requests] }
          end;
       Val -> { reply, Val, Requests}
    end;
loop_function(Requests, { Op, V1}) ->
    case search({Op, V1}, Requests) of
       false ->
          case Op of
	    square -> { reply, V1*V1, [{Op,V1,V1*V1} | Requests]}
	  end;
       Val -> { reply, Val, Requests }
    end.


search({Op,V1,V2}, [{Op,V1,V2,Res} | _Tail ]) -> Res;
search({Op,V1}, [{Op,V1,Res} | _Tail ]) -> Res;
search(Tuple, [_Head, Tail]) -> search(Tuple, Tail);
search(_Val, []) -> false;
search(_,_) -> false.

multiply(X,Y) ->
    genserver:request(server,{mul, X, Y}).

subtract(X,Y) ->
    genserver:request(server,{sub, X, Y}).

add(X,Y) ->
    genserver:request(server,{add, X, Y}).

square(X) ->
    genserver:request(server,{square, X}).
     

main_server() ->
    start_server(),
    
    Nums1 = [2,4,5,6],
    Nums2 = [3,5,7,9],
    Ops = [mul,sub,mul,add],
    RefsArgs = [{make_ref(), Num1, Num2, Op} || Num1 <- Nums1, Num2 <- Nums2, Op <- Ops],
    % Send the messages with the reference added to them
    [server ! { Op, Num1, Num2, self(), Ref } || {Ref, Num1, Num2, Op} <- RefsArgs],
    % Receive messages in the correct order
    Ans = [receive {answer, V1, V2, Op, An, Ref} -> { V1, V2, Op, An}  end || {Ref, _, _, _} <- RefsArgs],
    Print = fun(X) -> case X of mul -> "*"; add -> "+"; sub -> "-" end end,
    [ io:fwrite("~p ~s ~p = ~p~n", [V1,Print(Op),V2,An]) || {V1, V2, Op, An} <- Ans ].


main_genserver() ->
    start_genserver(),

    List = [add(2,3), multiply(2,3), subtract(3,1), add(2,3) ],
    [square(5) | List].



